import imp
import logging
import sys

from ..api.common import SessionHook


logger = logging.getLogger(__name__)
debug, warn = (logger.debug, logger.warn,)


class ScriptHost(object):
    """
    Plugin that provides the 'python' feature, emulating an environment for
    python code similar to the one provided by python-vim bindings.
    """
    def __init__(self, nvim):
        self.provides = ['python']
        self.nvim = nvim
        # context where all code will run
        self.module = imp.new_module('__main__')
        nvim.script_context = self.module
        # it seems some plugins assume 'sys' is already imported, so do it now
        exec('import sys', self.module.__dict__)
        sys.modules['vim'] = nvim.with_hook(LegacyEvalHook())

    def python_execute(self, script):
        exec(script, self.module.__dict__)

    def python_execute_file(self, file_path):
        with open(file_path) as f:
            script = compile(f.read(), file_path, 'exec')
            exec(script, self.module.__dict__)

    def python_do_range(self, start, stop, code):
        nvim = self.nvim
        start -= 1
        stop -= 1
        fname = '_vim_pydo'

        # Python3 code (exec) must be a string, mixing bytes with
        # function_def would use bytes.__repr__ instead
        if isinstance and isinstance(code, bytes):
            code = code.decode(nvim.options['encoding'].decode('ascii'))
        # define the function
        function_def = 'def %s(line, linenr):\n %s' % (fname, code,)
        exec(function_def, self.module.__dict__)
        # get the function
        function = self.module.__dict__[fname]
        while start <= stop:
            # Process batches of 5000 to avoid the overhead of making multiple
            # API calls for every line. Assuming an average line length of 100
            # bytes, approximately 488 kilobytes will be transferred per batch,
            # which can be done very quickly in a single API call.
            sstart = start
            sstop = min(start + 5000, stop)
            lines = nvim.current.buffer.get_slice(sstart, sstop, True, True)
            for i, line in enumerate(lines):
                linenr = i + sstart + 1
                result = str(function(line, linenr))
                if result:
                    lines[i] = result
            start = sstop + 1
            nvim.current.buffer.set_slice(sstart, sstop, True, True, lines)
        # delete the function
        del self.module.__dict__[fname]

    def python_eval(self, expr):
        return eval(expr, self.module.__dict__)


class LegacyEvalHook(SessionHook):

    """Injects legacy `vim.eval` behavior to a Nvim instance."""

    def __init__(self):
        super(LegacyEvalHook, self).__init__(from_nvim=self._string_eval)

    def _string_eval(self, obj, session, method, kind):
        if method == 'vim_eval' and isinstance(obj, (int, long, float)):
            return str(obj)
        return obj



"""API for working with Nvim buffers."""
from .common import Remote, RemoteMap
from ..compat import IS_PYTHON3


__all__ = ('Buffer')


if IS_PYTHON3:
    basestring = str


class Buffer(Remote):

    """A remote Nvim buffer."""

    def __init__(self, session, code_data):
        """Initialize from session and code_data immutable object.

        The `code_data` contains serialization information required for
        msgpack-rpc calls. It must be immutable for Buffer equality to work.
        """
        self._session = session
        self.code_data = code_data
        self.vars = RemoteMap(session, 'buffer_get_var', 'buffer_set_var',
                              self)
        self.options = RemoteMap(session, 'buffer_get_option',
                                 'buffer_set_option', self)

    def __len__(self):
        """Return the number of lines contained in a Buffer."""
        return self._session.request('buffer_get_length', self)

    def __getitem__(self, idx):
        """Get a buffer line or slice by integer index.

        Indexes may be negative to specify positions from the end of the
        buffer. For example, -1 is the last line, -2 is the line before that
        and so on.

        When retrieving slices, omiting indexes(eg: `buffer[:]`) will bring
        the whole buffer.
        """
        if not isinstance(idx, slice):
            return self._session.request('buffer_get_line', self, idx)
        include_end = False
        start = idx.start
        end = idx.stop
        if start is None:
            start = 0
        if end is None:
            end = -1
            include_end = True
        return self._session.request('buffer_get_slice', self, start, end,
                                     True, include_end)

    def __setitem__(self, idx, lines):
        """Replace a buffer line or slice by integer index.

        Like with `__getitem__`, indexes may be negative.

        When replacing slices, omiting indexes(eg: `buffer[:]`) will replace
        the whole buffer.
        """
        if not isinstance(idx, slice):
            if lines is None:
                return self._session.request('buffer_del_line', self, idx)
            else:
                return self._session.request('buffer_set_line', self, idx,
                                             lines)
        if lines is None:
            lines = []
        include_end = False
        start = idx.start
        end = idx.stop
        if start is None:
            start = 0
        if end is None:
            end = -1
            include_end = True
        return self._session.request('buffer_set_slice', self, start, end,
                                     True, include_end, lines)

    def __iter__(self):
        """Iterate lines of a buffer.

        This will retrieve all lines locally before iteration starts. This
        approach is used because for most cases, the gain is much greater by
        minimizing the number of API calls by transfering all data needed to
        work.
        """
        lines = self[:]
        for line in lines:
            yield line

    def get_slice(self, start, stop, start_incl, end_incl):
        """More flexible wrapper for retrieving slices."""
        return self._session.request('buffer_get_slice', self, start, stop,
                                     start_incl, end_incl)

    def set_slice(self, start, stop, start_incl, end_incl, lines):
        """More flexible wrapper for replacing slices."""
        return self._session.request('buffer_set_slice', self, start, stop,
                                     start_incl, end_incl, lines)

    def append(self, lines, index=-1):
        """Append a string or list of lines to the buffer."""
        if isinstance(lines, basestring):
            lines = [lines]
        return self._session.request('buffer_insert', self, index, lines)

    def mark(self, name):
        """Return (row, col) tuple for a named mark."""
        return self._session.request('buffer_get_mark', self, name)

    def range(self, start, end):
        """Return a `Range` object, which represents part of the Buffer."""
        return Range(self, start, end)

    @property
    def name(self):
        """Get the buffer name."""
        return self._session.request('buffer_get_name', self)

    @name.setter
    def name(self, value):
        """Set the buffer name. BufFilePre/BufFilePost are triggered."""
        return self._session.request('buffer_set_name', self, value)

    @property
    def valid(self):
        """Return True if the buffer still exists."""
        return self._session.request('buffer_is_valid', self)

    @property
    def number(self):
        """Get the buffer number."""
        return self._session.request('buffer_get_number', self)


class Range(object):
    def __init__(self, buffer, start, end):
        self._buffer = buffer
        self.start = start - 1
        self.end = end

    def __len__(self):
        return self.end - self.start

    def __getitem__(self, idx):
        if not isinstance(idx, slice):
            return self._buffer[self._normalize_index(idx)]
        start = self._normalize_index(idx.start)
        end = self._normalize_index(idx.stop)
        if start is None:
            start = self.start
        if end is None:
            end = self.end
        return self._buffer[start:end]

    def __setitem__(self, idx, lines):
        if not isinstance(idx, slice):
            self._buffer[self._normalize_index(idx)] = lines
            return
        start = self._normalize_index(idx.start)
        end = self._normalize_index(idx.stop)
        if start is None:
            start = self.start
        if end is None:
            end = self.end
        self._buffer[start:end] = lines

    def __iter__(self):
        for i in range(self.start, self.end):
            yield self._buffer[i]

    def append(self, lines, i=None):
        i = self._normalize_index(i)
        if i is None:
            i = self.end
        self._buffer.append(lines, i)

    def _normalize_index(self, index):
        if index is None:
            return None
        if index < 0:
            index = self.end - 1
        else:
            index += self.start
            if index >= self.end:
                index = self.end - 1
        return index



import inspect
import logging
import os
import os.path
import sys
from imp import find_module, load_module
from traceback import format_exc

from ..compat import IS_PYTHON3


logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warn,)


class RedirectStream(object):
    def __init__(self, redirect_handler):
        self.redirect_handler = redirect_handler

    def write(self, data):
        self.redirect_handler(data)

    def writelines(self, seq):
        self.redirect_handler('\n'.join(seq))


class PluginHost(object):
    """
    Class that transforms the python interpreter into a plugin host for
    Neovim. It takes care of discovering plugins and routing events/calls
    sent by Neovim to the appropriate handlers(registered by plugins)
    """
    def __init__(self, nvim, preloaded=[]):
        self.nvim = nvim
        self.method_handlers = {}
        self.event_handlers = {}
        self.discovered_plugins = list(preloaded)
        self.installed_plugins = []

    def __enter__(self):
        nvim = self.nvim
        info('install import hook/path')
        self.hook = path_hook(nvim)
        sys.path_hooks.append(self.hook)
        nvim.VIM_SPECIAL_PATH = '_vim_path_'
        sys.path.append(nvim.VIM_SPECIAL_PATH)
        info('redirect sys.stdout and sys.stderr')
        self.saved_stdout = sys.stdout
        self.saved_stderr = sys.stderr
        sys.stdout = RedirectStream(lambda data: nvim.out_write(data))
        sys.stderr = RedirectStream(lambda data: nvim.err_write(data))
        debug('installing plugins')
        self.install_plugins()
        return self

    def __exit__(self, type, value, traceback):
        for plugin in self.installed_plugins:
            if hasattr(plugin, 'on_teardown'):
                plugin.teardown()
        nvim = self.nvim
        info('uninstall import hook/path')
        sys.path.remove(nvim.VIM_SPECIAL_PATH)
        sys.path_hooks.remove(self.hook)
        info('restore sys.stdout and sys.stderr')
        sys.stdout = self.saved_stdout
        sys.stderr = self.saved_stderr

    def discover_plugins(self):
        loaded = set()
        for directory in discover_runtime_directories(self.nvim):
            for name in os.listdir(directory):
                if not name.startswith(b'nvim_'):
                    continue
                name = os.path.splitext(name)[0]
                if name in loaded:
                    continue
                loaded.add(name)
                try:
                    discovered = find_module(name, [directory])
                except:
                    err_str = format_exc(5)
                    warn('error while searching module %s: %s', name, err_str)
                    continue
                debug('discovered %s', name)
                try:
                    file, pathname, description = discovered
                    module = load_module(name, file, pathname, description)
                    for name, value in inspect.getmembers(module,
                                                          inspect.isclass):
                        if name.startswith('Nvim'):
                            self.discovered_plugins.append(value)
                    debug('loaded %s', name)
                except:
                    err_str = format_exc(5)
                    warn('error while loading module %s: %s', name, err_str)
                    continue
                finally:
                    file.close()

    def install_plugins(self):
        self.discover_plugins()
        nvim = self.nvim
        features = nvim.metadata['features']
        registered = set()
        for plugin_class in self.discovered_plugins:
            cls_name = plugin_class.__name__
            debug('inspecting class %s', plugin_class.__name__)
            try:
                plugin = plugin_class(self.nvim)
            except:
                err_str = format_exc(5)
                warn('constructor for %s failed: %s', cls_name, err_str)
                continue
            methods = inspect.getmembers(plugin, inspect.ismethod)
            debug('registering event handlers for %s', plugin_class.__name__)
            for method_name, method in methods:
                if not method_name.startswith('on_'):
                    continue
                # event handler
                # Store all handlers with bytestring keys, since thats how
                # msgpack will deserialize method names
                event_name = method_name[3:].encode('utf-8')
                debug('registering %s event handler', event_name)
                if event_name not in self.event_handlers:
                    self.event_handlers[event_name] = [method]
                else:
                    self.event_handlers[event_name].append(
                        method.__get__(plugin, plugin_class))

            if hasattr(plugin, 'provides') and plugin.provides:
                for feature_name in plugin.provides:
                    if feature_name in registered:
                        raise Exception('A plugin already provides %s' %
                                        feature_name)
                    for method_name in features[feature_name]:
                        # encode for the same reason as above
                        enc_name = method_name.encode('utf-8')
                        self.method_handlers[enc_name] = getattr(
                            # Python 3 attributes need to be unicode instances
                            # so use `method_name` here
                            plugin, method_name)
                    debug('registered %s as a %s provider',
                          plugin_class.__name__,
                          feature_name)
                    nvim.register_provider(feature_name)
                    registered.add(feature_name)
            self.installed_plugins.append(plugin)

    def search_handler_for(self, name):
        for plugin in self.installed_plugins:
            methods = inspect.getmembers(plugin, inspect.ismethod)
            for method_name, method in methods:
                if method_name == name:
                    return method

    def on_request(self, name, args):
        handler = self.method_handlers.get(name, None)
        if not handler:
            handler = self.search_handler_for(name)
            if handler:
                self.method_handlers[name] = handler
            else:
                msg = 'no method handlers for "%s" were found' % name
                debug(msg)
                raise Exception(msg)

        debug("running method handler for '%s %s'", name, args)
        rv = handler(*args)
        debug("method handler for '%s %s' returns: %s", name, args, rv)
        return rv

    def on_notification(self, name, args):
        handlers = self.event_handlers.get(name, None)
        if not handlers:
            debug("no event handlers registered for %s", name)
            return

        debug('running event handlers for %s', name)
        for handler in handlers:
            handler(*args)

    def run(self):
        self.nvim.session.run(self.on_request, self.on_notification)


# This was copied/adapted from nvim-python help
def path_hook(nvim):
    def _get_paths():
        return discover_runtime_directories(nvim)

    def _find_module(fullname, oldtail, path):
        idx = oldtail.find('.')
        if idx > 0:
            name = oldtail[:idx]
            tail = oldtail[idx+1:]
            fmr = find_module(name, path)
            module = load_module(fullname[:-len(oldtail)] + name, *fmr)
            return _find_module(fullname, tail, module.__path__)
        else:
            fmr = find_module(fullname, path)
            return load_module(fullname, *fmr)

    class VimModuleLoader(object):
        def __init__(self, module):
            self.module = module

        def load_module(self, fullname, path=None):
            return self.module

    class VimPathFinder(object):
        @classmethod
        def find_module(cls, fullname, path=None):
            try:
                return VimModuleLoader(
                    _find_module(fullname, fullname, path or _get_paths()))
            except ImportError:
                return None

        @classmethod
        def load_module(cls, fullname, path=None):
            return _find_module(fullname, fullname, path or _get_paths())

    def hook(path):
        if path == nvim.VIM_SPECIAL_PATH:
            return VimPathFinder
        else:
            raise ImportError

    return hook


def discover_runtime_directories(nvim):
    rv = []
    for path in nvim.list_runtime_paths():
        if not os.path.exists(path):
            continue
        path1 = os.path.join(path, b'pythonx')
        if IS_PYTHON3:
            path2 = os.path.join(path, b'python3')
        else:
            path2 = os.path.join(path, b'python2')
        if os.path.exists(path1):
            rv.append(path1)
        if os.path.exists(path2):
            rv.append(path2)
    return rv



"""Common code for event loop implementations."""
import logging
import signal
import threading


logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warn,)


# When signals are restored, the event loop library may reset SIGINT to SIG_DFL
# which exits the program. To be able to restore the python interpreter to it's
# default state, we keep a reference to the default handler
default_int_handler = signal.getsignal(signal.SIGINT)


class BaseEventLoop(object):

    """Abstract base class for all event loops.

    Event loops act as the bottom layer for Nvim sessions created by this
    library. They hide system/transport details behind a simple interface for
    reading/writing bytes to the connected Nvim instance.

    This class exposes public methods for interacting with the underlying
    event loop and delegates implementation-specific work to the following
    methods, which subclasses are expected to implement:

    - `_init()`: Implementation-specific initialization
    - `_connect_tcp(address, port)`: connect to Nvim using tcp/ip
    - `_connect_socket(path)`: Same as tcp, but use a UNIX domain socket or
      or named pipe.
    - `_connect_stdio()`: Use stdin/stdout as the connection to Nvim
    - `_connect_spawn(argv)`: Use the argument vector `argv` to spawn an
      embedded Nvim that has it's stdin/stdout connected to the event loop.
    - `_start_reading()`: Called after any of _connect_* methods. Can be used
      to perform any post-connection setup or validation.
    - `_send(data)`: Send `data`(byte array) to Nvim. The data is only
    - `_run()`: Runs the event loop until stopped or the connection is closed.
      calling the following methods when some event happens:
      actually sent when the event loop is running.
      - `_on_data(data)`: When Nvim sends some data.
      - `_on_signal(signum)`: When a signal is received.
      - `_on_error(message)`: When a non-recoverable error occurs(eg:
        connection lost)
    - `_stop()`: Stop the event loop
    - `_interrupt(data)`: Like `stop()`, but may be called from other threads
      this.
    - `_setup_signals(signals)`: Add implementation-specific listeners for
      for `signals`, which is a list of OS-specific signal numbers.
    - `_teardown_signals()`: Removes signal listeners set by `_setup_signals`
    """

    def __init__(self, transport_type, *args):
        """Initialize and connect the event loop instance.

        The only arguments are the transport type and transport-specific
        configuration, like this:

        >>> BaseEventLoop('tcp', '127.0.0.1', 7450)
        Traceback (most recent call last):
            ...
        AttributeError: 'BaseEventLoop' object has no attribute '_init'
        >>> BaseEventLoop('socket', '/tmp/nvim-socket')
        Traceback (most recent call last):
            ...
        AttributeError: 'BaseEventLoop' object has no attribute '_init'
        >>> BaseEventLoop('stdio')
        Traceback (most recent call last):
            ...
        AttributeError: 'BaseEventLoop' object has no attribute '_init'
        >>> BaseEventLoop('spawn', ['nvim', '--embed', '-u', 'NONE'])
        Traceback (most recent call last):
            ...
        AttributeError: 'BaseEventLoop' object has no attribute '_init'

        This calls the implementation-specific initialization
        `_init`, one of the `_connect_*` methods(based on `transport_type`)
        and `_start_reading()`
        """
        self._transport_type = transport_type
        self._signames = dict((k, v) for v, k in signal.__dict__.items()
                              if v.startswith('SIG'))
        self._on_data = None
        self._error = None
        self._init()
        getattr(self, '_connect_{0}'.format(transport_type))(*args)
        self._start_reading()

    def connect_tcp(self, address, port):
        """Connect to tcp/ip `address`:`port`. Delegated to `_connect_tcp`."""
        info('Connecting to TCP address: %s:%d', address, port)
        self._connect_tcp(address, port)

    def connect_socket(self, path):
        """Connect to socket at `path`. Delegated to `_connect_socket`."""
        info('Connecting to %s', path)
        self._connect_socket(path)

    def connect_stdio(self):
        """Connect using stdin/stdout. Delegated to `_connect_stdio`."""
        info('Preparing stdin/stdout for streaming data')
        self._connect_stdio()

    def connect_spawn(self, argv):
        """Connect a new Nvim instance. Delegated to `_connect_spawn`."""
        info('Spawning a new nvim instance')
        self._connect_spawn(argv)

    def send(self, data):
        """Queue `data` for sending to Nvim."""
        debug("Sending '%s'", data)
        self._send(data)

    def interrupt(self):
        """Stop the event loop from another thread."""
        debug('Interrupted event loop from thread %s',
              threading.current_thread())
        self._interrupt()

    def run(self, data_cb):
        """Run the event loop."""
        if self._error:
            err = self._error
            if isinstance(self._error, KeyboardInterrupt):
                # KeyboardInterrupt is not destructive(it may be used in
                # the REPL).
                # After throwing KeyboardInterrupt, cleanup the _error field
                # so the loop may be started again
                self._error = None
            raise err
        self._on_data = data_cb
        self._setup_signals([signal.SIGINT, signal.SIGTERM])
        debug('Entering event loop')
        self._run()
        debug('Exited event loop')
        self._teardown_signals()
        signal.signal(signal.SIGINT, default_int_handler)
        self._on_data = None

    def stop(self):
        """Stop the event loop."""
        self._stop()
        debug('Stopped event loop')

    def _on_signal(self, signum):
        msg = 'Received {0}'.format(self._signames[signum])
        debug(msg)
        if signum == signal.SIGINT and self._transport_type == 'stdio':
            # When the transport is stdio, we are probably running as a Nvim
            # child process. In that case, we don't want to be killed by
            # ctrl+C
            return
        cls = Exception
        if signum == signal.SIGINT:
            cls = KeyboardInterrupt
        self._error = cls(msg)
        self.stop()

    def _on_error(self, error):
        debug(error)
        self._error = Exception(error)
        self.stop()

    def _on_interrupt(self):
        self.stop()



"""Msgpack handling in the event loop pipeline."""
import logging
from collections import deque

from msgpack import Packer, Unpacker


logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warn,)


class MsgpackStream(object):

    """Two-way msgpack stream that wraps a event loop byte stream.

    This wraps the event loop interface for reading/writing bytes and
    exposes an interface for reading/writing msgpack documents.
    """

    def __init__(self, event_loop):
        """Wrap `event_loop` on a msgpack-aware interface."""
        self._event_loop = event_loop
        self._posted = deque()
        self._packer = Packer(use_bin_type=True)
        self._unpacker = Unpacker()
        self._message_cb = None
        self._stopped = False

    def post(self, msg):
        """Post `msg` to the read queue of the `MsgpackStream` instance.

        Use the event loop `interrupt()` method to push msgpack objects from
        other threads.
        """
        self._posted.append(msg)
        self._event_loop.interrupt()

    def send(self, msg):
        """Queue `msg` for sending to Nvim."""
        debug('sent %s', msg)
        self._event_loop.send(self._packer.pack(msg))

    def run(self, message_cb):
        """Run the event loop to receive messages from Nvim.

        While the event loop is running, `message_cb` will be called whenever
        a message has been successfully parsed from the input stream.
        """
        self._message_cb = message_cb
        self._run()
        self._message_cb = None

    def stop(self):
        """Stop the event loop."""
        self._stopped = True
        self._event_loop.stop()

    def _run(self):
        self._stopped = False
        while not self._stopped:
            if self._posted:
                self._message_cb(self._posted.popleft())
                continue
            self._event_loop.run(self._on_data)

    def _on_data(self, data):
        self._unpacker.feed(data)
        while True:
            try:
                debug('waiting for message...')
                msg = next(self._unpacker)
                debug('received message: %s', msg)
                self._message_cb(msg)
            except StopIteration:
                debug('unpacker needs more data...')
                break



import platform
import sys

from setuptools import setup

install_requires = [
    'msgpack-python',
]

if sys.version_info < (3, 4):
    # trollius is just a backport of 3.4 asyncio module
    install_requires.append('trollius')

if not platform.python_implementation() == 'PyPy':
    # pypy already includes an implementation of the greenlet module
    install_requires.append('greenlet')

setup(name='neovim',
      version='0.0.20',
      description='Python client to neovim',
      url='http://github.com/neovim/python-client',
      download_url='https://github.com/neovim/python-client/archive/0.0.20.tar.gz',
      author='Thiago de Arruda',
      author_email='tpadilha84@gmail.com',
      license='MIT',
      packages=['neovim', 'neovim.api', 'neovim.msgpack_rpc',
                'neovim.msgpack_rpc.event_loop', 'neovim.plugins'],
      install_requires=install_requires,
      zip_safe=False)



"""Code shared between the API classes."""


class Remote(object):

    """Base class for Nvim objects(buffer/window/tabpage).

    Each type of object has it's own specialized class with API wrappers around
    the msgpack-rpc session. This implements equality which takes the remote
    object handle into consideration.
    """

    def __eq__(self, other):
        """Return True if `self` and `other` are the same object."""
        return (hasattr(other, 'code_data') and
                other.code_data == self.code_data)


class RemoteMap(object):

    """Represents a string->object map stored in Nvim.

    This is the dict counterpart to the `RemoteSequence` class, but it is used
    as a generic way of retrieving values from the various map-like data
    structures present in Nvim.

    It is used to provide a dict-like API to vim variables and options.
    """

    def __init__(self, session, get_method, set_method, self_obj=None):
        """Initialize a RemoteMap with session, getter/setter and self_obj."""
        self._get = _wrap(session, get_method, self_obj)
        self._set = None
        if set_method:
            self._set = _wrap(session, set_method, self_obj)

    def __getitem__(self, key):
        """Return a map value by key."""
        return self._get(key)

    def __setitem__(self, key, value):
        """Set a map value by key(if the setter was provided)."""
        if not self._set:
            raise TypeError('This dict is read-only')
        self._set(key, value)

    def __delitem__(self, key):
        """Delete a map value by associating None with the key."""
        if not self._set:
            raise TypeError('This dict is read-only')
        return self._set(key, None)

    def __contains__(self, key):
        """Check if key is present in the map."""
        try:
            self._get(key)
            return True
        except:
            return False


class RemoteSequence(object):

    """Represents a sequence of objects stored in Nvim.

    This class is used to wrap msgapck-rpc functions that work on Nvim
    sequences(of lines, buffers, windows and tabpages) with an API that
    is similar to the one provided by the python-vim interface.

    For example, the 'buffers' property of the `Nvim class is a RemoteSequence
    sequence instance, and the expression `nvim.buffers[0]` is translated to
    session.request('vim_get_buffers')[0].

    It can also receive an optional self_obj that will be passed as first
    argument of the request. For example, `tabpage.windows[0]` is translated
    to: session.request('tabpage_get_windows', tabpage_instance)[0].

    One important detail about this class is that all methods will fetch the
    sequence into a list and perform the necessary manipulation
    locally(iteration, indexing, counting, etc).
    """

    def __init__(self, session, method, self_obj=None):
        """Initialize a RemoteSequence with session, method and self_obj."""
        self._fetch = _wrap(session, method, self_obj)

    def __len__(self):
        """Return the length of the remote sequence."""
        return len(self._fetch())

    def __getitem__(self, idx):
        """Return a sequence item by index."""
        if not isinstance(idx, slice):
            return self._fetch()[idx]
        return self._fetch()[idx.start:idx.stop]

    def __iter__(self):
        """Return an iterator for the sequence."""
        items = self._fetch()
        for item in items:
            yield item

    def __contains__(self, item):
        """Check if an item is present in the sequence."""
        for i in self._fetch():
            if i._handle == item._handle:
                return True
        return False


class SessionHook(object):

    """Pair of functions to filter objects coming/going from/to Nvim.

    Filter functions receive the following arguments:

    - obj: The object to process
    - session: The current session object
    - method: The method name
    - kind: Kind of filter, can be one of:
        - 'request' for requests coming from Nvim
        - 'notification' for notifications coming from Nvim
        - 'out-request' for requests going to Nvim

    Whatever is returned from the function is used as a replacement for `obj`.

    This class also provides a `compose` method for composing hooks.
    """

    identity = lambda obj, session, method, kind: obj

    def __init__(self, from_nvim=identity, to_nvim=identity):
        """Initialize a SessionHook with from/to filters."""
        self.from_nvim = from_nvim
        self.to_nvim = to_nvim

    def compose(self, other):
        """Compose two SessionHook instances.

        This works by composing the individual from/to filters and creating
        a new SessionHook instance with the composed filters.
        """
        def comp(f1, f2):
            if f1 is SessionHook.identity:
                return f2
            if f2 is SessionHook.identity:
                return f1
            return lambda o, s, m, k: f1(f2(o, s, m, k), s, m, k)

        return SessionHook(comp(other.from_nvim, self.from_nvim),
                           comp(other.to_nvim, self.to_nvim))


class DecodeHook(SessionHook):

    """SessionHook subclass that decodes utf-8 strings coming from Nvim.

    This class is useful for python3, where strings are now unicode by
    default(byte strings need to be prefixed with "b").
    """

    def __init__(self, encoding='utf-8', encoding_errors='strict'):
        """Initialize with encoding and encoding errors policy."""
        self.encoding = encoding
        self.encoding_errors = encoding_errors
        super(DecodeHook, self).__init__(from_nvim=self._decode_if_bytes)

    def _decode_if_bytes(self, obj, session, method, kind):
        if isinstance(obj, bytes):
            return obj.decode(self.encoding, errors=self.encoding_errors)
        return obj


class SessionFilter(object):

    """Wraps a session-like object with a SessionHook instance.

    This class can be used as a drop-in replacement for a sessions, the
    difference is that a hook is applied to all data passing through a
    SessionFilter instance.
    """

    def __init__(self, session, hook):
        """Initialize with a Session(or SessionFilter) and a hook.

        If `session` is already a SessionFilter, it's hook will be extracted
        and composed with `hook`.
        """
        if isinstance(session, SessionFilter):
            self._hook = session._hook.compose(hook)
            self._session = session._session
        else:
            self._hook = hook
            self._session = session
        # Both filters are applied to `walk` so objects are transformed
        # recursively
        self._in = self._hook.from_nvim
        self._out = self._hook.to_nvim

    def post(self, name, *args):
        """Wrapper for Session.post."""
        self._session.post(name, *args)

    def next_message(self):
        """Wrapper for Session.next_message."""
        msg = self._session.next_message()
        return walk(self._in, msg, self, msg[1], msg[0])

    def request(self, name, *args):
        """Wrapper for Session.request."""
        args = walk(self._out, args, self, name, 'out-request')
        return walk(self._in, self._session.request(name, *args), self, name,
                    'out-request')

    def run(self, request_cb, notification_cb):
        """Wrapper for Session.run."""
        def filter_request_cb(name, args):
            result = request_cb(self._in(name, self, name, 'request'),
                                walk(self._in, args, self, name, 'request'))
            return walk(self._out, result, self, name, 'request')

        def filter_notification_cb(name, args):
            notification_cb(self._in(name, self, name, 'notification'),
                            walk(self._in, args, self, name, 'notification'))

        self._session.run(filter_request_cb, filter_notification_cb)

    def stop(self):
        """Wrapper for Session.stop."""
        self._session.stop()


def walk(fn, obj, *args):
    """Recursively walk an object graph applying `fn`/`args` to objects."""
    if type(obj) in [list, tuple]:
        return list(walk(fn, o, *args) for o in obj)
    if type(obj) is dict:
        return dict((walk(fn, k, *args), walk(fn, v, *args)) for k, v in
                    obj.items())
    return fn(obj, *args)


def _wrap(session, method, self_obj):
    if self_obj is not None:
        return (lambda *args: session.request(method, self_obj, *args))
    else:
        return (lambda *args: session.request(method, *args))



"""Python client for Nvim.

Client library for talking with Nvim processes via it's msgpack-rpc API.
"""
import logging
import os

from .api import DecodeHook, Nvim, SessionHook
from .msgpack_rpc import (socket_session, spawn_session, stdio_session,
                          tcp_session)
from .plugins import PluginHost, ScriptHost


__all__ = ('tcp_session', 'socket_session', 'stdio_session', 'spawn_session',
           'start_host', 'DecodeHook', 'Nvim', 'SessionHook')


def start_host(session=None):
    """Promote the current process into python plugin host for Nvim.

    Start msgpack-rpc event loop for `session`, listening for Nvim requests
    and notifications. It registers Nvim commands for loading/unloading
    python plugins.

    The sys.stdout and sys.stderr streams are redirected to Nvim through
    `session`. That means print statements probably won't work as expected
    while this function doesn't return.

    This function is normally called at program startup and could have been
    defined as a separate executable. It is exposed as a library function for
    testing purposes only.
    """
    logger = logging.getLogger(__name__)
    if 'NVIM_PYTHON_LOG_FILE' in os.environ:
        logfile = os.environ['NVIM_PYTHON_LOG_FILE'].strip()
        handler = logging.FileHandler(logfile, 'w')
        handler.formatter = logging.Formatter(
            '%(asctime)s [%(levelname)s @ '
            '%(filename)s:%(funcName)s:%(lineno)s] %(process)s - %(message)s')
        logging.root.addHandler(handler)
        level = logging.INFO
        if 'NVIM_PYTHON_LOG_LEVEL' in os.environ:
            l = getattr(logging,
                        os.environ['NVIM_PYTHON_LOG_LEVEL'].strip(),
                        level)
            if isinstance(l, int):
                level = l
        logger.setLevel(level)
    if not session:
        session = stdio_session()
    nvim = Nvim.from_session(session)
    with PluginHost(nvim, preloaded=[ScriptHost]) as host:
        host.run()


# Required for python 2.6
class NullHandler(logging.Handler):
    def emit(self, record):
        pass


if not logging.root.handlers:
    logging.root.addHandler(NullHandler())
